from yta_multimedia.video.edition.effect.moviepy.moviepy_effect import MoviepyEffect
from moviepy.editor import vfx, VideoFileClip, CompositeVideoClip, ImageClip
from typing import Union


class FadeInMoviepyEffect(MoviepyEffect):
    """
    This effect will make the video appear progressively
    lasting the provided 'duration' time or the whole 
    clip time duration if None 'duration' provided.

    The 'color' provided must be an array containing
    the rgb colors (default is [0, 0, 0]).
    """
    __MOVIEPY_EFFECT_NAME = 'fadein'
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], duration: float = None, color = None):
        self.__clip = clip

        if duration is None:
            duration = self.__clip.duration

        if color is None:
            color = [0, 0, 0]

        self.__parameters['duration'] = duration
        self.__parameters['initial_color'] = color

    def get_moviepy_vfx_effect(self):
        return getattr(vfx, self.__MOVIEPY_EFFECT_NAME, None)
    
    def process_parameters(self):
        if not self.__parameters['duration']:
            self.__parameters['duration'] = self.__clip.duration
        else:
            if self.__parameters['duration'] > self.__clip.duration:
                self.__parameters['duration'] = self.__clip.duration
            elif self.__parameters['duration'] <= 0:
                self.__parameters['duration'] = 0

        if not self.__parameters['initial_color']:
            self.__parameters['initial_color'] = [0, 0, 0]
        # TODO: Check color is a valid value.

        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        return self.__clip.fx(self.get_moviepy_vfx_effect(), **self.process_parameters())



from moviepy.editor import VideoFileClip, concatenate_videoclips
from yta_general_utils.checker.type import variable_is_type
from typing import Union


def concatenate_moviepy(videoclips: list[VideoFileClip], output_filename: Union[str, None] = None):
    """
    Concatenates the provided moviepy 'videoclips' into a new one.

    If 'output_filename' provided, the new clip be stored with that
    name and that name.

    This method will return the new clips array.
    """
    if not videoclips:
        return None
    
    if variable_is_type(output_filename, str):
        if not output_filename:
            return None

    clip = concatenate_videoclips(videoclips)

    if output_filename:
        clip.write_videofile(output_filename)

    return clip

# TODO: This is being tested
def concatenate_images_ffmpeg(abspath, output_filename = 'merged.mp4'):
    # TODO: Do it
    from yta_general_utils.file_processor import list
    from subprocess import run
    # TODO: Check that abspath is a folder
    # TODO: Make this able to accept images array (?)
    images = list(abspath)

    text = ''
    for image in images:
        text += "file '" + image + "'\n"

    filename = 'append_videos.txt'
    #output_filename = abspath + '/' + output_filename
    output_filename = abspath + output_filename
    with open(filename, 'w') as the_file:
        the_file.write(text)

    # https://www.reddit.com/r/ffmpeg/comments/ks8zfs/comment/gieu7x6/?utm_source=share&utm_medium=web3x&utm_name=web3xcss&utm_term=1&utm_content=share_button
    # https://stackoverflow.com/questions/38368105/ffmpeg-custom-sequence-input-images/51618079#51618079
    # https://stackoverflow.com/a/66014158

    # TODO: Output needs to be an abspath or will be in images folder
    # ffmpeg -i %d.png -vcodec png z.mov   This is said in stackoverflow
    #command = 'ffmpeg -f concat -y -safe 0 -i ' + filename + ' -r 60 ' + output_filename
    #command = 'ffmpeg -f concat -i ' + filename + ' -r 60 ' + output_filename
    #command = 'ffmpeg -f concat -safe 0 -y -i ' + filename + ' -r 60 -vcodec png ' + output_filename
    #command = 'ffmpeg -f concat -safe 0 -y -i ' + filename + ' -vcodec png ' + output_filename
    #command = 'ffmpeg -f concat -safe 0 -y -i ' + filename + ' -r 60 -c:v prores -pix_fmt yuva444p10le ' + output_filename
    command = 'ffmpeg -f concat -safe 0 -y -r 60  -i ' + filename + ' -c:v qtrle -pix_fmt argb ' + output_filename
    run(command)



from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.utils.dimensions import fitting_text, manim_width_to_width
from yta_multimedia.video.generation.manim.constants import SCENE_WIDTH
from manim import *


class TextWordByWordManimAnimation(BaseManimAnimation):
    """
    The provided 'text' is shown word by word in the center of the scene
    with a fixed width.
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, text = 'texto de prueba, ¿te gusta?', duration: float = 2, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid.
        """
        # Check and validate all parameters
        parameters = {}

        if super().parameter_is_mandatory('text', self.required_parameters) and not text:
            raise Exception('Field "text" is mandatory. Aborting manim creation...')
        
        parameters['text'] = text

        if super().parameter_is_mandatory('duration', self.required_parameters) and not duration:
            raise Exception('Field "duration" is mandatory. Aborting manim creation...')
        if duration < 0 or duration > 100:
            raise Exception('Field "duration" value is not valid. Must be between 0 and 100')
        
        parameters['duration'] = duration

        if not output_filename:
            output_filename = 'output.mov'

        # Generate the animation when parameters are valid
        super().generate(parameters, output_filename = output_filename)

        return output_filename

    def animate(self):
        words = self.parameters['text'].split(' ')
        word_duration = float(self.parameters['duration']) / len(words)
        for word in words:
            text = fitting_text(word, manim_width_to_width(SCENE_WIDTH / 6))
            text = Text(word, font_size = text.font_size, stroke_width = 2.0, font = 'Arial').shift(DOWN * 0)
            self.add(text)
            self.wait(word_duration)
            self.remove(text)

    


from yta_multimedia.video.edition.effect.moviepy.moviepy_effect import MoviepyEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip
from typing import Union
from PIL import Image

import numpy as np
import math


class ZoomOutMoviepyEffect(MoviepyEffect):
    """
    This effect will zoom out the clip, on the center.

    TODO: This effect is not smooth as it makes it have
    a black border. Maybe removing it (?)
    """
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], zoom_ratio = None):
        self.__clip = clip

        if zoom_ratio is None:
            zoom_ratio = 0.01

        self.__parameters['zoom_ratio'] = zoom_ratio

    def process_parameters(self):
        if not self.__parameters['zoom_ratio']:
            self.__parameters['zoom_ratio'] = 0.01
        else:
            # Zoom is by now limited to [0.01 - 0.2] ratio
            if self.__parameters['zoom_ratio'] > 0.2:
                self.__parameters['zoom_ratio'] = 0.2
            elif self.__parameters['zoom_ratio'] <= 0.01:
                self.__parameters['zoom_ratio'] = 0.01

        return self.__parameters
    
    def __effect_calculations(self, get_frame, t, zoom_ratio):
        # TODO: This is not a good effect because it doesn't 
        # work perfectly... Maybe removing it
        img = Image.fromarray(get_frame(t))
        base_size = tuple(1.4*x for x in img.size)

        new_size = [
            math.ceil(img.size[0] * (1 - (zoom_ratio * t))),
            math.ceil(img.size[1] * (1 - (zoom_ratio * t)))
        ]

        # The new dimensions must be even.
        new_size[0] = new_size[0] + (new_size[0] % 2)
        new_size[1] = new_size[1] + (new_size[1] % 2)
        print(new_size)

        img = img.resize(new_size, Image.LANCZOS)

        x = math.ceil((new_size[0] - base_size[0]) / 2)
        y = math.ceil((new_size[1] - base_size[1]) / 2)

        img = img.crop([
            x, y, new_size[0] - x, new_size[1] - y
        ]).resize(base_size, Image.LANCZOS)

        result = np.array(img)
        img.close()

        return result
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        return self.__clip.fl(lambda get_frame, t: self.__effect_calculations(get_frame, t, **self.process_parameters()))



from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.resources import get_resource
from yta_general_utils.text.transformer import remove_accents
from manim import *


class MagazineTextIsWrittenManimAnimation(BaseManimAnimation):
    """
    Makes an animation in which the provided 'text' appears
    with special magazine characters through an animation
    as if they were written. It is only one row of text 
    limited to 30 characters.
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, text: str = 'Texto', duration: float = 2, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid. The 
        'text' parameter is limited to 30 characters.
        """
        # Check and validate all parameters
        parameters = {}

        if super().parameter_is_mandatory('text', self.required_parameters) and not text:
            raise Exception('Field "text" is mandatory. Aborting manim creation...')
        
        if len(text) > 30:
            raise Exception('Field "text" has a maximum of 30 characters.')
        
        parameters['text'] = text

        if super().parameter_is_mandatory('duration', self.required_parameters) and not duration:
            raise Exception('Field "duration" is mandatory. Aborting manim creation...')
        if duration < 0 or duration > 100:
            raise Exception('Field "duration" value is not valid. Must be between 0 and 100')
        
        parameters['duration'] = duration

        if not output_filename:
            output_filename = 'output.mov'

        # Generate the animation when parameters are valid
        super().generate(parameters, output_filename = output_filename)

        return output_filename
    
    def animate(self):
        """
        This code will generate the manim animation and belongs to the
        Scene manim object.
        """
        DISTANCE_BETWEEN_LETTERS = 0.30

        text = remove_accents(self.parameters['text'])

        elements = []
        # TODO: Make the text fit attending to how long it is
        for char in text:
            elements_len = len(elements)
            new_letter = self.__get(char)

            if elements_len > 0:
                new_letter.move_to(elements[elements_len - 1]).shift(RIGHT * DISTANCE_BETWEEN_LETTERS)

            elements.append(new_letter)

        all_mobjects = Group(*elements)
        all_mobjects.move_to((0, 0, 0)) # We place all elements in the center

        # We need to make the animation fit the 
        self.wait(1 / 60)

        # If we have a 'duration' of 3 seconds that means 180 FPS.
        # We will be using (1 / 60) before and after the loop so
        # we substract it. We will make one animation per each 
        # mobject, so thats how this work dynamically
        EACH_LETTER_WAITING_TIME = self.parameters['duration'] * 60 / len(all_mobjects) - (2 / 60)
        for mobject in all_mobjects:
            self.add(mobject)
            self.wait(EACH_LETTER_WAITING_TIME / 60)

        self.wait(1 / 60)

        # TODO: If this doesn't work we just set our own animation duration and thats all
        # TODO: We need to make the static alternativa in which the text is just added
        # TODO: This is the Animation, what about creating a big Mobject (?)

    # TODO: This below is repeated in the other magazine scene
    def __get(self, letter: str = 'a'):
        if not letter:
            return None
        
        letter = letter[0:1].lower()
        mobject_group = None

        if letter == 'a':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1p_NrqZEf3lhQl_lB_PMwgI4FNIbVkSs-/view?usp=sharing', 0.2, 0)
        elif letter == 'b':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1KO0WsYNcocowFeG7Uql_71dn4tlEyUEJ/view?usp=sharing', 0.2, 0)
        elif letter == 'c':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1pFC5l3C0JylftXsDajI2hYJwgFgHFHeX/view?usp=sharing', 0.2, 0)
        elif letter == 'd':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1f7MPrRmczO-oaQYG5tdadg0-q7X8Avm9/view?usp=sharing', 0.2, 0)
        elif letter == 'e':
            mobject_group = self.__get_letter(letter, '#325e2e', 'https://drive.google.com/file/d/1JLb2dPK5AdGqk3MdF1KzKwQnisU1fYwf/view?usp=sharing', 0.2, 0)
        elif letter == 'f':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1QA2u8ppuZ9nWFlbfyOBDP3sZ0GRRAcRD/view?usp=sharing', 0.2, 0)
        elif letter == 'g':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1-3PlY9N-X83gyYQWX-M-t0R_gUkp8TMp/view?usp=sharing', 0.2, 0)
        elif letter == 'h':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1_e5SzuadQUzMia_KhiLa2p9qntxg-kIC/view?usp=sharing', 0.2, 0)
        elif letter == 'i':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1pWGKVyDGyXcLX2Xwg8pxAV0FHSchB9_J/view?usp=sharing', 0.2, 0)
        elif letter == 'j':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/17fLo90TghQIRsrLvIm7gRIwKjJkcNgun/view?usp=sharing', 0.2, 0)
        elif letter == 'k':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1yoLkzH1o8vZ4RjVCDWza6WCWvZ8y5MxN/view?usp=sharing', 0.2, 0)
        elif letter == 'l':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1cZAncrpB0j8FD64opEzQdyeCtQVJbajU/view?usp=sharing', 0.2, 0)
        elif letter == 'm':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1XQ02p-veaiSWLyx2PozCfK2QhflDCnpu/view?usp=sharing', 0.2, 0)
        elif letter == 'n':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1gp-9ZLAVVPeOO5uitF6zVOJqeK6443R8/view?usp=sharing', 0.2, 0)
        # TODO: Make 'ñ'
        elif letter == 'o':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/13H8nmkH9UiTn7vG57WgpC0v5S_UNATPM/view?usp=sharing', 0.2, 0)
        elif letter == 'p':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1YNSFeSAujWWFuwrqU_A-iw_Xk-l2vXvU/view?usp=sharing', 0.2, 0)
        elif letter == 'q':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1k4MuXPS9M2H4Z3BCnCf64yXcBwf6QEKO/view?usp=sharing', 0.2, 0)
        elif letter == 'r':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1nM_jNdzGGV60Z8fAjJp-Q43RXCmwsqh0/view?usp=sharing', 0.2, 0)
        elif letter == 's':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1bJC8IAMsHOdqeVxdPBhuV3_KYIE0DZSZ/view?usp=sharing', 0.2, 0)
        elif letter == 't':
            mobject_group = self.__get_letter(letter, WHITE, 'https://drive.google.com/file/d/1H4t4iS_yrsAqg7QEdAhxwDOOahBh54dl/view?usp=sharing', 0.2, 0)
        elif letter == 'u':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1M_uMgx9CBkwM0tKTMH_YiyGRiYPIxbXa/view?usp=sharing', 0.2, 0)
        elif letter == 'v':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/12SGuNyNdF03ua2Ro7GIoAf7ETUB0jDWl/view?usp=sharing', 0.2, 0)
        elif letter == 'w':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1aDQ0vhhrnh-BzVgtnTv9HOFSm7cgxycc/view?usp=sharing', 0.2, 0)
        elif letter == 'x':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1wjCkyibwEYz3G19XZiO9viGkNhFRgY0M/view?usp=sharing', 0.2, 0)
        elif letter == 'y':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1aIEu0EP8a-tDHY6aGvAnFGtGyFBn9M7D/view?usp=sharing', 0.2, 0)
        elif letter == 'z':
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1k7EryiaWPX8HRlnB7tJtEL1eI85z7KD_/view?usp=sharing', 0.2, 0)
        # Special ones
        elif letter == '¿': # Background is the same as '?'
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1cGm0r4cdDi4b4ZSS_VxH3yJdyJ5sGB6Q/view?usp=sharing', 0.2, 0)
        elif letter == '?': # Background is the same as '¿'
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1cGm0r4cdDi4b4ZSS_VxH3yJdyJ5sGB6Q/view?usp=sharing', 0.2, 0)
        elif letter == '¡': # Background is the same as '!'
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1J4aL2yf1V_lCjbxzgXE1HyoXoyBRVXwx/view?usp=sharing', 0.2, 0)
        elif letter == '!': # Backgroudn is the same '¡'
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1J4aL2yf1V_lCjbxzgXE1HyoXoyBRVXwx/view?usp=sharing', 0.2, 0)
        # General ones (those who I don't specify manually)
        # elif letter == ' ':
        #     mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1fLol_Fe_AiQ6ZDY8pWW6yBXSFg9PnTb_/view?usp=sharing', 0.2, 0)
        else:   # Any other sign, even the space
            mobject_group = self.__get_letter(letter, BLACK, 'https://drive.google.com/file/d/1fLol_Fe_AiQ6ZDY8pWW6yBXSFg9PnTb_/view?usp=sharing', 0.2, 0)

        return mobject_group

    def __get_letter(self, letter: str = 'a', color: ManimColor = WHITE, google_drive_url: str = None, background_scale: float = 0.2, background_shift: float = 0):
        """
        Generates the default Mobject Group for our own characters
        """
        if not letter:
            return None

        if not google_drive_url:
            return None

        letter = letter[0:1]

        if letter == ' ':
            return Group(Text('a', font_size = 48, color = color).set_opacity(0))

        TMP_FILENAME = get_resource(google_drive_url, 'resources/manim/magazine_letters/' + letter + '.svg')
        letter_background = SVGMobject(TMP_FILENAME).scale(0.2).shift(UP * background_shift)
        letter_text = Text(letter, font_size = 32, color = color)
        letter_group = Group(letter_background, letter_text)
        
        return letter_group


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.constants import MANDATORY_CONFIG_PARAMETER
from yta_general_utils.downloader.google_drive import download_file_from_google_drive
from yta_general_utils.temp import create_temp_filename
from manim import *


class SimpleTextManimAnimation(BaseManimAnimation):
    # TODO: Maybe make this more strict with 'type' also
    required_parameters = {
        'text': MANDATORY_CONFIG_PARAMETER,
        'duration': MANDATORY_CONFIG_PARAMETER,
    }

    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, text: str = 'Texto', duration: float = 2, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid.
        """
        # Check and validate all parameters
        parameters = {}

        if super().parameter_is_mandatory('text', self.required_parameters) and not text:
            raise Exception('Field "text" is mandatory. Aborting manim creation...')
        
        parameters['text'] = text

        if super().parameter_is_mandatory('duration', self.required_parameters) and not duration:
            raise Exception('Field "duration" is mandatory. Aborting manim creation...')
        if duration < 0 or duration > 100:
            raise Exception('Field "duration" value is not valid. Must be between 0 and 100')
        
        parameters['duration'] = duration

        if not output_filename:
            output_filename = 'output.mov'

        # Generate the animation when parameters are valid
        super().generate(parameters, output_filename = output_filename)

        return output_filename
    
    def animate_one(self):
        """
        This code will generate the manim animation and belongs to the
        Scene manim object.
        """
        text = Text(self.parameters['text'], font_size = 140, stroke_width = 2.0, font = 'Haettenschweiler').shift(DOWN * 0).scale(0.001)
        self.wait(1 / 60)
        self.add(text)
        self.add_sound(download_file_from_google_drive('https://drive.google.com/file/d/1WPS8uWB1LTuzPzxQ2Zcp1FpwvLM3fhM5/view?usp=sharing', create_tmp_filename('tmp.mp3')))
        self.play(text.animate.scale(1000), run_time = 49 / 60)
        self.play(Rotate(text, 0.03), run_time = 3 / 60)
        self.play(Rotate(text, -0.04), run_time = 3 / 60)
        self.play(Rotate(text, -0.02), run_time = 3 / 60)
        self.play(Rotate(text, 0.04), run_time = 3 / 60)
        self.play(Rotate(text, -0.01), run_time = 3 / 60)
        self.play(Rotate(text, 0.03), run_time = 3 / 60)
        self.play(Rotate(text, -0.04), run_time = 3 / 60)
        self.play(Rotate(text, -0.02), run_time = 3 / 60)
        self.play(Rotate(text, 0.04), run_time = 3 / 60)
        self.play(Rotate(text, -0.01), run_time = 3 / 60)
        #self.add_sound(TOUSE_ABSOLUTE_PATH + 'sounds/xp_error.mp3')
        #self.play(AddTextLetterByLetter(text), run_time = self.parameters['duration'])
        
        #simple_play_animation(self, Write, text, self.parameters['duration'])

    def animate_two(self):
        """
        This code will generate the manim animation and belongs to the
        Scene manim object.
        """
        text = Text(self.parameters['text'], font_size = 140, stroke_width = 2.0, font = 'Haettenschweiler').shift(DOWN * 0).scale(7 / 10)
        self.wait(1 / 60)
        self.add(text)
        self.play(text.animate.scale(10 / 7), run_time = 6 / 60)
        self.play(ApplyWave(text), run_time = 30 / 60)

    def animate(self):
        text = Text(self.parameters['text'], font_size = 26, font = 'Minecraftia').shift(DOWN * 0).scale(1)
        self.add(text)
        self.add_sound(download_file_from_google_drive('https://drive.google.com/file/d/1Dzeb6Qae4UdpmuA6U9d6t3MXnvhzukzg/view?usp=sharing', create_tmp_filename('tmp.mp3')))
        self.wait(self.parameters['text'])


from yta_general_utils.file.writer import write_file
from yta_general_utils.programming.path import get_project_abspath

import json


CONFIG_MANIM_ABSPATH = get_project_abspath() + 'manim_parameters.json'

def write_manim_config_file(json_data):
    """
    Writes in the configuration file that we use to share
    parameters with manim software. This is the way to 
    share parameters to the process.
    
    TODO: I would like to be able to handle manim through 
    python code directly and not an external process I run,
    but for now this is working.
    """
    # We serialize json to str
    json_object_str = json.dumps(json_data, indent = 4)
    
    write_file(json_object_str, CONFIG_MANIM_ABSPATH)

def read_manim_config_file():
    """
    Reads the configuration file and returns it as a json
    object.
    """
    json_object = None

    # TODO: Create a general method to read files (?)
    with open(CONFIG_MANIM_ABSPATH, 'r') as config_file:
        json_object = json.load(config_file)
        config_file.close()

    return json_object


from pedalboard import Pedalboard, Reverb
from math import trunc
from yta_general_utils.file.filename import get_file_extension
from yta_general_utils.file.checker import file_is_audio_file
from yta_general_utils.audio.converter import mp3_to_wav
from yta_general_utils.temp import create_temp_filename

import numpy as np
import soundfile as sf


def slow_and_reverb_audio_file(audio_filename: str, output_filename: str, room_size: float = 0.75, damping: float = 0.5, wet_level: float = 0.08, dry_level: float = 0.2, delay: float = 2, slow_factor: float = 0.08):
    # Extracted from here: https://github.com/samarthshrivas/LoFi-Converter-GUI
    # But there is no only one: https://github.com/topics/slowedandreverbed
    if not audio_filename:
        return None
    
    if not file_is_audio_file(audio_filename):
        return None

    if not output_filename:
        return None
    
    if get_file_extension(audio_filename) != '.wav':
        # TODO: Handle other formats, by now I think it is .mp3 only
        tmp_filename = create_temp_filename('transformed_audio.wav')
        mp3_to_wav(audio_filename, tmp_filename)
        audio_filename = tmp_filename

    audio, sample_rate = sf.read(audio_filename)
    sample_rate -= trunc(sample_rate * slow_factor)

    # Adding reverb effect
    reverved_board = Pedalboard([
        Reverb(
            # TODO: I need to learn more about these parameters
            room_size = room_size,
            damping = damping,
            wet_level = wet_level,
            dry_level = dry_level
        )
    ])

    # Adding other surrounding effects
    audio_with_effects = reverved_board(audio, sample_rate)
    channel_1 = audio_with_effects[:, 0]
    channel_2 = audio_with_effects[:, 1]
    shift_length = delay * 1000
    shifted_channel_1 = np.concatenate((np.zeros(shift_length), channel_1[:-shift_length]))
    combined_signal = np.hstack((shifted_channel_1.reshape(-1, 1), channel_2.reshape(-1, 1)))

    # Write the slowed and reverved output file
    sf.write(output_filename, combined_signal, sample_rate)

    return output_filename


from yta_multimedia.video.edition.effect.moviepy.moviepy_effect import MoviepyEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, CompositeAudioClip, AudioFileClip, concatenate_videoclips
from yta_multimedia.resources.video.effect.sound.drive_urls import PHOTO_GOOGLE_DRIVE_DOWNLOAD_URL
from yta_multimedia.video.edition.effect.moviepy.blink_moviepy_effect import BlinkMoviepyEffect
from yta_multimedia.resources import get_resource
from yta_multimedia.video.edition.effect.constants import EFFECTS_RESOURCES_FOLDER
from typing import Union


class PhotoMoviepyEffect(MoviepyEffect):
    """
    TODO: Add description
    """
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip]):
        self.__clip = clip

    def process_parameters(self):
        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        if not self.__clip:
            return None
        
        self.process_parameters()

        # Change this parameter to adjust the effect
        effect_duration = 0.2
        if self.__clip.duration < effect_duration:
            effect_duration = self.__clip.duration

        TMP_FILENAME = get_resource(PHOTO_GOOGLE_DRIVE_DOWNLOAD_URL, EFFECTS_RESOURCES_FOLDER + 'sounds/photo_taken.mp3')

        if self.__clip.duration == effect_duration:
            # If clip is shorter than our default effect duration time, do it with
            # the clip duration
            self.__clip = BlinkMoviepyEffect(self.__clip, effect_duration, [255, 255, 255]).apply()
        else:
            # We force the effect to be 'effect_duration' seconds longer and in the
            # middle of the provided clip
            half_duration = self.__clip.duration / 2
            half_effect_duration = effect_duration / 2
            self.__clip = concatenate_videoclips([
                self.__clip.subclip(0, half_duration - effect_duration),
                BlinkMoviepyEffect(self.__clip.subclip(half_duration - half_effect_duration, half_duration + half_effect_duration), effect_duration, [255, 255, 255]).apply(),
                self.__clip.subclip(half_duration + half_effect_duration, self.__clip.duration)
            ])
        
        self.__clip.audio = CompositeAudioClip([
            self.__clip.audio,
            AudioFileClip(TMP_FILENAME).set_duration(effect_duration)
        ])

        return self.__clip


from yta_general_utils.checker.type import variable_is_type
from yta_general_utils.file.checker import file_exists, file_is_video_file
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.file.writer import write_file
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip
from typing import Union
from subprocess import run


def generate_videoclip_from_image(image_filename: Union[ImageClip, str], duration: float = 1, output_filename: Union[str, None] = None):
    """
    Receives an image as 'image_filename' and creates an ImageClip of
    'duration' seconds. It will be also stored as a file if 
    'output_filename' is provided.

    # TODO: Should this method go into 'video.utils' instead of here (?)
    """
    if not image_filename:
        return None
    
    if duration <= 0:
        return None
    
    if not duration:
        return None
    
    if variable_is_type(output_filename, str):
        if not output_filename:
            return None
    
    if variable_is_type(image_filename, str):
        # ADV: By now we are limiting this to 60 fps
        image_filename = ImageClip(image_filename).set_fps(60).set_duration(duration)

    if output_filename:
        image_filename.write_videofile(output_filename)

    return image_filename

def parse_parameter_as_moviepy_clip(video: Union[VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, str]):
    """
    Checks if the provided 'video' is a valid moviepy video, of any 
    type, and it is returned. If it is a string, it is check if the
    filename string is a valid video file and its loaded. Anything
    unexpected will raise an Exception.

    This method returns a the video as a moviepy clip.
    """
    if not video:
        raise Exception('No "video" provided.')
    
    if not isinstance(video, (VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, str)):
        raise Exception('The "video" parameter provided is not a VideoFileClip nor a CompositeVideoClip nor an ImageClip nor a ColorClip nor a string.')
    
    if isinstance(video, str):
        if not file_exists(video):
            raise Exception('The "video" parameter provided file does not exist.')
        
        if not file_is_video_file(video):
            raise Exception('The "video" parameter provided file is not a valid video file.')
        
        video = VideoFileClip(video)

    return video

def concatenate_videos_ffmpeg(videos_abspath, output_abspath: str):
    """
    This method concatenates the videos provided in 'videos_abspath'
    and builds a new video, stored locally as 'output_abspath'.

    This method uses ffmpeg to concatenate the videos, so they must
    have the same resolution.
    """
    text = ''
    for video_abspath in videos_abspath:
        text += f'file \'{video_abspath}\'\n'

    filename = create_temp_filename('append_videos.txt')
    write_file(text, filename)

    # TODO: Make a custom call from python not as command
    command = 'ffmpeg -f concat -safe 0 -i ' + filename + ' -c copy ' + output_abspath
    run(command)
    return VideoFileClip(output_abspath)


def zoom_in_t_func(zoom_ratio: float = 0.2):
    """
    Function that returns the resize value to make a moviepy
    video be zoomed in the 'zoom_ratio' zoom all the time.
    """
    return 1 + zoom_ratio

def zoom_out_t_func(zoom_ratio: float = 0.2):
    """
    Function that returns the resize value to make a moviepy
    video be zoomed out the 'zoom_ratio' zoom all the time.
    """
    return 1 - zoom_ratio

def linear_zoom_in_t_func(t, duration, zoom_ratio: float = 0.2):
    """
    Function that returns the resize value to make a moviepy
    video zoom in linearly to the 'zoom_ratio' zoom.
    """
    # TODO: Check 'zoom_ratio' is valid

    return 1 + zoom_ratio * (t / duration)
    """
    if t < 4:
        return 1 + 0.2 * t  # Zoom-in.
    elif 4 <= t <= 6:
        return 1 + 0.2 * 4  # Stay.
    else: # 6 < t
        return 1 + 0.2 * (duration - t)  # Zoom-out.
    """

def linear_zoom_out_t_func(t, duration, zoom_ratio: float = 0.2):
    """
    Function that returns the resize value to make a moviepy
    video zoom out linearly to the 'zoom_ratio' zoom.
    """
    # TODO: Check 'zoom_ratio' is valid

    return 1 + zoom_ratio * (duration - (t / duration))




from yta_multimedia.video.edition.effect.moviepy.position.static.circles_at_position_moviepy_effect import CirclesAtPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.static.stay_at_position_moviey_effect import StayAtPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.static.shake.shake_at_position_moviey_effect import ShakeAtPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.static.shake.shake_increasing_at_position_moviey_effect import ShakeIncreasingAtPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.static.shake.shake_decreasing_at_position_moviey_effect import ShakeDecreasingAtPositionMoviepyEffect


from yta_multimedia.image.edition.filter.workers.sketch.util import *
from random import *


def sortlines(lines):
    clines = lines[:]
    slines = [clines.pop(0)]
    while clines != []:
        x,s,r = None,1000000,False
        for l in clines:
            d = distsum(l[0],slines[-1][-1])
            dr = distsum(l[-1],slines[-1][-1])
            if d < s:
                x,s,r = l[:],d,False
            if dr < s:
                x,s,r = l[:],s,True

        clines.remove(x)
        if r == True:
            x = x[::-1]
        slines.append(x)
    return slines

def visualize(lines):
    import turtle
    wn = turtle.Screen()
    t = turtle.Turtle()
    t.speed(0)
    t.pencolor('red')
    t.pd()
    for i in range(0,len(lines)):
        for p in lines[i]:
            t.goto(p[0]*640/1024-320,-(p[1]*640/1024-320))
            t.pencolor('black')
        t.pencolor('red')
    turtle.mainloop()

if __name__=="__main__":
    import linedraw
    #linedraw.draw_hatch = False
    lines = linedraw.sketch("Lenna")
    #lines = sortlines(lines)
    visualize(lines)


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from yta_multimedia.video.edition.effect.utils.resize_t_functions import linear_zoom_out_t_func
from moviepy.editor import CompositeVideoClip


class LinearZoomOutVideoEffect(VideoEffect):
    """
    Creates a linear Zoom out effect in the provided video.
    """
    def apply(self, zoom_ratio: float = None):
        fps = self.video.fps
        duration = self.video.duration
        screensize = self.video.size

        # TODO: Check that 'zoom_ratio' is valid
        if zoom_ratio is None:
            zoom_ratio = 0.2

        effected_video = (
            self.video
            .resize(screensize)
            .resize(lambda t: linear_zoom_out_t_func(t, duration, zoom_ratio))
            .set_position(('center', 'center'))
            .set_duration(duration)
            .set_fps(fps)
        )

        return CompositeVideoClip([effected_video], size = screensize)


    


from abc import ABC, abstractmethod


class MoviepyEffect(ABC):
    """
    Abstract class to be inherited by all my custom moviepy effects
    so I can control they belong to this family.

    A moviepy effect (or what I call like that) is an effect that is
    applied directly to the video by using only the moviepy editor
    and/or moviepy vfx module. It could be a simple moviepy effect
    made an object to simplify the work with it, or a more complex
    effect that is build with some different small effects.
    """
    @abstractmethod
    def process_parameters(self):
        pass
    
    @abstractmethod
    def apply(self):
        pass



from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_center import CoordinateCenter
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_corner import CoordinateCorner
from yta_multimedia.video.edition.effect.moviepy.position.enums import ScreenPosition
from yta_multimedia.video.edition.effect.moviepy.position.utils.position import get_moviepy_position
from yta_multimedia.video.edition.effect.moviepy.position.utils.move import sinusoidal_movement
from yta_general_utils.file.checker import file_is_video_file
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip
from typing import Union


class MoveSinusoidPositionMoviepyEffect(BasePositionMoviepyEffect):
    """
    Effect of moving the element from one place in the screen (or outside
    of bounds) to another, linearly.
    """
    def apply(self, initial_position: Union[ScreenPosition, CoordinateCorner, CoordinateCenter] = ScreenPosition.OUT_LEFT, final_position: Union[ScreenPosition, CoordinateCorner, CoordinateCenter] = ScreenPosition.CENTER):
        """
        Applies the effect to the 'video' provided when initializing this
        effect class, and puts the video over a static black background
        image of 1920x1080.
        """
        background_video = ColorClip((1920, 1080), [0, 0, 0], duration = self.video.duration)

        return self.apply_over_video(background_video, initial_position, final_position)

    def apply_over_video(self, background_video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], initial_position: Union[ScreenPosition, CoordinateCorner, CoordinateCenter] = ScreenPosition.OUT_LEFT, final_position: Union[ScreenPosition, CoordinateCorner, CoordinateCenter] = ScreenPosition.CENTER):
        """
        TODO: Write well

        Applies the effect on the video used when instantiating the
        effect, but applies the effect by placing it over the 
        'background_video' provided in this method (the 
        'background_video' will act as a background video for the 
        effect applied on the initial video).

        This method will set the video used when instantiating the
        effect as the most important, and its duration will be 
        considered as that. If the 'background_video' provided 
        has a duration lower than the original video, we will
        loop it to reach that duration. If the video is shorter
        than the 'background_video', we will crop the last one
        to fit the original video duration.
        """
        if not background_video:
            raise Exception('No "background_video" provided.')
        
        if isinstance(background_video, str):
            if not file_is_video_file:
                raise Exception('Provided "background_video" is not a valid video file.')
            
            background_video = VideoFileClip(background_video)

        if not isinstance(initial_position, ScreenPosition) and not isinstance(initial_position, CoordinateCenter) and not isinstance(initial_position, CoordinateCorner):
            raise Exception('Provided "initial_position" is not a valid ScreenPosition, CoordinateCenter or CoordinateCorner.')

        if not isinstance(final_position, ScreenPosition) and not isinstance(final_position, CoordinateCenter) and not isinstance(final_position, CoordinateCorner):
            raise Exception('Provided "final_position" is not a valid ScreenPosition, CoordinateCenter or CoordinateCorner.')

        background_video = super().process_background_video(background_video)

        initial_position = get_moviepy_position(self.video, background_video, initial_position)
        final_position = get_moviepy_position(self.video, background_video, final_position)

        effect = self.video.set_position(lambda t: sinusoidal_movement(t, initial_position, final_position, self.video.duration, self.video.fps)).set_start(0).set_duration(self.video.duration)
        
        return CompositeVideoClip([
            background_video,
            effect
        ])


from yta_multimedia.video.edition.effect.moviepy.moviepy_effect import MoviepyEffect
from moviepy.editor import vfx, VideoFileClip, CompositeVideoClip, ImageClip
from typing import Union


class FadeOutMoviepyEffect(MoviepyEffect):
    """
    This effect will make the video disappear 
    progressively lasting the provided 'duration' 
    time or the whole clip time duration if None
    'duration' provided.
    """
    __MOVIEPY_EFFECT_NAME = 'fadeout'
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], duration: float = None, color = None):
        self.__clip = clip

        if duration is None:
            duration = self.__clip.duration

        if color is None:
            color = [0, 0, 0]

        self.__parameters['duration'] = duration
        self.__parameters['final_color'] = color

    def get_moviepy_vfx_effect(self):
        return getattr(vfx, self.__MOVIEPY_EFFECT_NAME, None)
    
    def process_parameters(self):
        if not self.__parameters['duration']:
            self.__parameters['duration'] = self.__clip.duration
        else:
            if self.__parameters['duration'] > self.__clip.duration:
                self.__parameters['duration'] = self.__clip.duration
            elif self.__parameters['duration'] <= 0:
                self.__parameters['duration'] = 0

        if not self.__parameters['final_color']:
            self.__parameters['final_color'] = [0, 0, 0]
        # TODO: Check color is a valid value.

        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        return self.__clip.fx(self.get_moviepy_vfx_effect(), **self.process_parameters())



from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.move import MoveLinearPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.static import StayAtPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.enums import ScreenPosition
from yta_general_utils.file.checker import file_is_video_file
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, concatenate_videoclips
from typing import Union
from random import randrange


class SlideRandomPositionMoviepyEffect(BasePositionMoviepyEffect):
    """
    Effect of appearing from TOP, TOP_LEFT, BOTTOM, RIGHT, etc. 
    staying at the center, and dissapearing from the opposite 
    edge. This animation will spend 1/6 of the time in the 
    entrance, 4/6 of the time staying at the center, and 1/6 of 
    the time in the exit.
    """
    def apply_over_video(self, background_video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        This effect will make the 'self.video' appear from outside
        of the screen (from the 'in_screen_edge'), will stay in
        the middle of the screen for 4/6 times of its duration, and
        will go away through the 'out_screen_edge' edge of the 
        screen. All over the provided 'background_video'.

        Applies the effect on the video used when instantiating the
        effect, but applies the effect by placing it over the 
        'background_video' provided in this method (the 
        'background_video' will act as a background video for the 
        effect applied on the initial video).

        This method will set the video used when instantiating the
        effect as the most important, and its duration will be 
        considered as that. If the 'background_video' provided 
        has a duration lower than the original video, we will
        loop it to reach that duration. If the video is shorter
        than the 'background_video', we will crop the last one
        to fit the original video duration.
        """
        if not background_video:
            raise Exception('No "background_video" provided.')
        
        if isinstance(background_video, str):
            if not file_is_video_file:
                raise Exception('Provided "background_video" is not a valid video file.')
            
            background_video = VideoFileClip(background_video)

        background_video = super().process_background_video(background_video)

        random_position = ScreenPosition.in_and_out_positions_as_list()

        movement_time = background_video.duration / 6
        stay_time = background_video.duration / 6 * 4

        effect = concatenate_videoclips([
            MoveLinearPositionMoviepyEffect(self.video.subclip(0, movement_time)).apply_over_video(background_video.subclip(0, movement_time), random_position[0], ScreenPosition.CENTER),
            StayAtPositionMoviepyEffect(self.video.subclip(movement_time, movement_time + stay_time)).apply_over_video(background_video.subclip(movement_time, movement_time + stay_time), ScreenPosition.CENTER),
            MoveLinearPositionMoviepyEffect(self.video.subclip(movement_time + stay_time, self.video.duration)).apply_over_video(background_video.subclip(movement_time + stay_time, self.video.duration), ScreenPosition.CENTER, random_position[1])
        ])

        return effect


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.classes.loading_bar.parameter_classes.loading_bar_image import LoadingBarImage
from yta_multimedia.video.generation.manim.classes.loading_bar.mobjects.loading_bar_mobject import LoadingBarMobject
from yta_multimedia.video.generation.manim.utils.dimensions import manim_height_to_height
from manim import linear


class LoadingBarWithImagesManimAnimation(BaseManimAnimation):
    """
    Loading bar with images. It will go from 'start_percentage'
    to 'end_percentage' in the provided 'duration' time. It will
    show the bar progressing and also the 'images' if provided.

    Images will be displayed and will be static or in movement.
    If image 'start_percentage' and 'end_percentage' is the same,
    it will be static at that percentage position. If they are
    different, it will move from 'start_percentage' to
    'end_percentage' in the animation 'duration' time.

    You can put the same 'start_percentage' and 'end_percentage'
    in one image than in the loading bar to make it advance
    at the same time the progress bar does.

    Here is an example of 'images' object you can pass:
    loading_bar_images = [
        LoadingBarImage('minecraft_sword_128x128.png', 60, 0, 100),
        LoadingBarImage('arrow_up_128x128.png', -60, 0, 100),
        LoadingBarImage('minecraft_sword_128x128.png', 60, 40, 40),
        LoadingBarImage('minecraft_sword_128x128.png', 60, 78, 78)
    ]
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, start_percentage: int = 0, end_percentage: int = 100, duration: float = 5, images: list[LoadingBarImage] = None, output_filename: str = None):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid.
        """
        # Check and validate all parameters
        parameters = {}

        # TODO: By now I will avoid using 'required_parameters' and use the
        # logic by itself just here:
        # start_percentage is mandatory
        # end_percentage is mandatory
        # duration is mandatory
        # images is optional

        # Mandatory parameters
        if start_percentage == None or start_percentage < 0:
            raise Exception('"start_percentage" parameter must be a valid int between [0, 100]')
        
        if end_percentage == None or end_percentage < 0:
            raise Exception('"end_percentage" parameter must be a valid int between [0, 100]')
        
        if duration == None or duration <= 0 or duration > 120:
            raise Exception('"duration" parameter must be a valid float between (0, 120]')
        
        parameters['start_percentage'] = start_percentage
        parameters['end_percentage'] = end_percentage
        parameters['duration'] = duration

        # Optional parameters
        parameters['images'] = []
        if images and len(images) > 0:
            # We need to pass the information as JSON to be written
            # TODO: Check 'images' parameter is valid
            for image in images:
                parameters['images'].append(image.toJSON())

        if not output_filename:
            output_filename = 'output.mov'

        # Generate the animation when parameters are valid
        super().generate(parameters, output_filename = output_filename)

        return output_filename

    def animate(self):
        loading_bar = LoadingBarMobject()
        
        # We make LoadingBarImages objects again
        images = []
        for image in self.parameters['images']:
            images.append(LoadingBarImage(image['image_filename'], manim_height_to_height(image['y']), image['start_percentage'], image['end_percentage']))

        loading_bar_animation = loading_bar.get_animation(images, self.parameters['start_percentage'], self.parameters['end_percentage'], self.parameters['duration'])

        # We add the mobjects to forze them to appear
        self.add(*loading_bar.get_mobjects())

        self.play(*loading_bar_animation, rate_func = linear, run_time = self.parameters['duration'])



from yta_general_utils.programming.env import get_current_project_env
from PIL import Image

import requests
import io


class HugginFace:
    """
    Represents the functionality with the HugginFace.co platform that
    has a lot of AI models and API endpoints. Make sure you have the
    'HUGGINFACE_API_KEY' in your .env file, instantiate this class
    with your desired API endpoint, send the payload you need and
    make the magic.

    You go to any existing model in HugginFace platform, then click on
    'Deploy' (on the right side) and 'Inference API' and there you have
    the url and the query structure.
    """
    API_KEY = get_current_project_env("HUGGINFACE_API_KEY")
    HEADERS = {"Authorization": f"Bearer {API_KEY}"}
    
    def __init__(self, api_url):
        self.api_url = api_url

    def __request(self, payload = None, data = None):
        response = requests.post(self.api_url, headers = self.HEADERS, data = data, json = payload)

        return response

    def request_content(self, payload = None, data = None):
        """
        Makes a request with the provided 'payload' and returns
        the '.content' information returned.
        """
        return self.__request(payload = payload, data = data).content

    def request_json(self, payload = None, data = None):
        """
        Makes a request with the provided 'payload' and returns
        the '.json()' information returned.
        """
        return self.__request(payload = payload, data = data).json()
    
    def generate_image(self, payload, output_filename = None):
        """
        Makes a request with the provided payload and stores the 
        image received as response in the local storage as the
        provided 'output_filename'. This method must be used with
        image generation urls.

        This method will store locally the image if 'output_filename'
        provided, and will always return the image read with PIL.

        @param
            **payload**
            The information you need to send in the payload. It 
            depends on the endpoint you are using. It could be 
            just a string or maybe a json (dict). You must check
            it.

        @param
            **output_filename**
            The filename in which you want to store the received
            image if you want to store it. If None provided, it
            wont be stored.
        """
        image_content = self.request_content(payload = payload)

        if output_filename:
            with open(output_filename, 'wb') as outfile:
                outfile.write(image_content)

        return Image.open(io.BytesIO(image_content))













# from PIL import Image

# import requests
# import os
# import io

# API_KEY = os.environ.get("HUGGINFACE_API_KEY")
# HEADERS = {"Authorization": f"Bearer {API_KEY}"}



# def __request(api_url, payload):
#     response = requests.post(api_url, headers = HEADERS, json = payload)
#     print(response)
#     return response

# def __request_image(api_url, payload):
#     return __request(api_url, payload).content

# def __request_json(api_url, payload):
#     return __request(api_url, payload).json()

# def test_image():
#     #API_URL = 'https://api-inference.huggingface.co/models/stabilityai/stable-cascade'
#     #API_URL = 'https://api-inference.huggingface.co/models/stabilityai/stable-cascade-prior'
#     #API_URL = 'https://api-inference.huggingface.co/models/runwayml/stable-diffusion-v1-5'
#     #API_URL = 'https://api-inference.huggingface.co/models/stabilityai/stable-diffusion-2-1'
#     API_URL = 'https://api-inference.huggingface.co/models/prompthero/openjourney'

#     PAYLOAD = {
#         'inputs': 'Hiperrealistic monk between two cars in the middle of a city'
#     }

#     response = __request_image(API_URL, PAYLOAD)
#     Image.open(io.BytesIO(response)).save('test_tintin.png')

# def test_tintin():
#     API_URL = 'https://api-inference.huggingface.co/models/Pclanglais/TintinIA'

#     PAYLOAD = {
#         'inputs': 'Astronaut riding a horse and holding a coffee'
#     }

#     response = __request_image(API_URL, PAYLOAD)
#     Image.open(io.BytesIO(response)).save('test_tintin.png')

# def test_gemma():
#     API_URL = 'https://api-inference.huggingface.co/models/google/gemma-7b'

#     PAYLOAD = {
#         'inputs': '¿Cuántos habitantes tiene España?'
#     }

#     response = __request_json(API_URL, PAYLOAD)

#     print(response)

# def test():
#     API_URL = 'https://api-inference.huggingface.co/models/deepset/roberta-base-squad2'
    
#     PAYLOAD = {
#         'inputs': {
#             'question': 'What is my name?',
#             'context': 'My name is Clara and I live in Berkeley.',
#         }
#     }

#     response = __request_json(API_URL, PAYLOAD)
#     #response has, in this case, 'answer'
#     print(response)


from yta_general_utils.file.checker import file_is_audio_file
from yta_general_utils.temp import create_temp_filename
from moviepy.editor import AudioFileClip, CompositeAudioClip
from pydub import AudioSegment
from typing import Union

import numpy as np
import io
import scipy.io.wavfile as wavfile


# TODO: Refactor this, please
def append_audio_clip_to_audio_clip(base_audio_clip: AudioFileClip, new_audio_clip: AudioFileClip, start_time: float = 0.0, output_filename: str = None):
    """
    We receive a 'base_audio_clip' in which we will append the 'new_audio_clip'
    starting at 'start_time' seconds since the first one started. A combined 
    audio will be generated and returned.

    If 'output_filename' is provided, it will be also written in that 
    destination.
    """
    if start_time < 0:
        start_time = 0

    combined = CompositeAudioClip([base_audio_clip, new_audio_clip.set_start(start_time)])
    # TODO: Why 441000 here below?
    combined.fps = 44100

    if output_filename:
        combined.write_audiofile(output_filename)

    return combined

# TODO: Refactor this, please
def append_audio_file_to_audio_file(base_audio_filename: str, new_audio_filename: str, start_time: float = 1.0, output_filename: str = None):
    """
    We receive an audio file called 'base_audio_filename' and we append the provided
    'new_audio_filename' to that existing audio to sound over it. That second 
    sound will start in the 'start_time' second of the first 'audio_filename'.

    If 'output_filename' provided, it will write the new combined audio. If not,
    it will only return the new CompositeAudioClip object.

    # TODO: Is this method really necessary (?) I think it is not...
    """
    if not base_audio_filename:
        return None
    
    if not file_is_audio_file(base_audio_filename):
        return None
    
    if not new_audio_filename:
        return None
    
    if not file_is_audio_file(new_audio_filename):
        return None
    
    if start_time < 0:
        return None

    return append_audio_clip_to_audio_clip(AudioFileClip(base_audio_filename), AudioFileClip(new_audio_filename), start_time, output_filename)


def audio_to_audiosegment(audio):
    """
    Forces the provided 'audio' to be a pydub AudioSegment
    and returns it if valid 'audio' provided or raises an
    Exception if not.
    """
    if not audio:
        raise Exception('No "audio" provided.')
    
    if isinstance(audio, str):
        if not file_is_audio_file(audio):
            raise Exception('Provided "audio" filename is not a valid audio file.')
        audio = AudioSegment.from_file(audio)
    elif isinstance(audio, np.ndarray):
        # TODO: Check this
        audio = numpy_to_audiosegment(audio)
    elif isinstance(audio, AudioFileClip):
        audio = audiofileclip_to_audiosegment(audio)
    elif isinstance(audio, CompositeAudioClip):
        audio = compositeaudioclip_to_audiosegment(audio)

    return audio

def audio_to_audiofileclip(audio):
    """
    Forces the provided 'audio' to be a moviepy AudioFileClip
    and returns it if valid 'audio' provided or raises an
    Exception if not.
    """
    if not audio:
        raise Exception('No "audio" provided.')
    
    if isinstance(audio, str):
        if not file_is_audio_file(audio):
            raise Exception('Provided "audio" filename is not a valid audio file.')
        audio = AudioFileClip(audio)
    elif isinstance(audio, np.ndarray):
        # TODO: Check this works
        audio = numpy_to_audiofileclip(audio)
    elif isinstance(audio, AudioSegment):
        audio = audiosegment_to_audiofileclip(audio)

    return audio

def audiosegment_to_audiofileclip(audio: AudioSegment):
    """
    This method reads the provided AudioSegment as a buffer and converts
    it to a moviepy AudioFileClip without writting any file.

    TODO: Please, make it through memory and not writting files.
    """
    if not audio:
        raise Exception('No "audio" provided.')
    
    if not isinstance(audio, AudioSegment):
        raise Exception('The "audio" parameter provided is not an AudioSegment.')
    
    # TODO: I have not been able to create an AudioFileClip dinamically
    # from memory information. I don't want to write but...
    tmp_filename = create_temp_filename('tmp_audio.wav')
    audio.export(tmp_filename, format = 'wav')

    return AudioFileClip(tmp_filename)

def audiofileclip_to_audiosegment(audio: AudioFileClip):
    """
    This method returns the provided 'audio' moviepy AudioFileClip converted
    to a pydub AudioSegment.

    TODO: This method currently writes a temporary file to make the 
    conversion. This needs to be improved to avoid writting files.
    """
    if not audio:
        raise Exception('No "audio" provided.')
    
    if not isinstance(audio, AudioFileClip):
        raise Exception('The "audio" parameter provided is not an AudioFileClip.')

    # TODO: Please, improve this to be not writting files
    tmp_filename = create_temp_filename('tmp_audio.wav')
    audio.write_audiofile(tmp_filename)
    audio = AudioSegment.from_file(tmp_filename, format = 'wav')

    return audio

def compositeaudioclip_to_audiosegment(audio: CompositeAudioClip):
    """
    This method returns the provided 'audio' moviepy CompositeAudioClip
    converted to a pydub AudioSegment.

    TODO: This method currently writes a temporary file to make the 
    conversion. This needs to be improved to avoid writting files.
    """
    if not audio:
        raise Exception('No "audio" provided.')
    
    if not isinstance(audio, CompositeAudioClip):
        raise Exception('The "audio" parameter provided is not an CompositeAudioClip.')
    
    # TODO: Please, improve this to be not writting files
    tmp_filename = create_temp_filename('tmp_audio.wav')
    audio.write_audiofile(tmp_filename)
    audio = AudioSegment.from_file(tmp_filename, format = 'wav')

    return audio

# TODO: This has not been tested properly
def numpy_to_audiosegment(audio, sample_rate):
    """
    Convers the provided 'audio' numpy array, that contains the audio data
    and must be in float32 or int16, to a pydub AudioSegment.
    """
    # Normalize audio_array if it's not already in int16 format
    if audio.dtype != np.int16:
        if audio.dtype != np.float32:
            raise Exception('The "audio" parameter provided is not np.int16 nor np.float32.')
        
        # Assuming the audio_array is in float32 with values between -1 and 1
        audio = (audio * 32767).astype(np.int16)
    
    # Create a BytesIO object
    with io.BytesIO() as buffer:
        wavfile.write(buffer, sample_rate, audio)
        buffer.seek(0)
        
        audio_segment = AudioSegment.from_file(buffer, format = 'wav')
    
    return audio_segment

# TODO: This method has not been tested properly.
def audiosegment_to_numpy(audio: AudioSegment):
    """
    This method turns the provided 'audio' AudioSegment into a numpy
    array by converting it first to an AudioFileClip and then to a
    numpy.

    TODO: Please, maybe it is a better (more direct) way

    TODO: This method has not been tested properly
    """
    if not audio:
        raise Exception('No "audio" provided.')
    
    if not isinstance(audio, AudioSegment):
        raise Exception('The "audio" parameter provided is not an AudioSegment.')
    
    # TODO: Maybe this is not the best way, I need to test and improve this
    return np.array(audio.get_array_of_samples())
    #return audiofileclip_to_numpy(audiosegment_to_audiofileclip(audio))


# TODO: This has not been tested properly
def audiofileclip_to_numpy(audio: Union[AudioFileClip]):
    """
    Convers the provided 'audio' moviepy AudioFileClip to a numpy
    array that will be np.float32.
    """
    if not audio:
        raise Exception('No "audio" provided.')
    
    if not isinstance(audio, AudioFileClip):
        raise Exception('The "audio" parameter provided is not a moviepy AudioFileClip.')

    chunk_size = 5*1024*1024
    audio_chunks = []
    for chunk in audio.iter_chunks(chunksize = chunk_size):
        # Convertir cada fragmento a un array numpy y añadirlo a la lista
        audio_array = np.frombuffer(chunk, dtype=np.int16)
        
        # Normalizar si el audio es estéreo (tendría dos columnas)
        if len(audio_array) > 0 and len(audio_array) % 2 == 0:
            audio_array = audio_array.reshape(-1, 2)
        
        audio_chunks.append(audio_array)
    
    # Concatenar todos los fragmentos en un solo array
    full_audio_array = np.concatenate(audio_chunks, axis = 0)
    
    # Convertir a float32 y normalizar
    full_audio_array = full_audio_array.astype(np.float32)
    if np.max(np.abs(full_audio_array)) > 1.0:
        full_audio_array /= np.max(np.abs(full_audio_array))
    
    return full_audio_array

# TODO: This has not been tested properly
def numpy_to_audiofileclip(audio: np.ndarray):
    if not audio:
        raise Exception('No "audio" provided.')
    
    if not isinstance(audio, np.ndarray):
        # TODO: Check this works
        raise Exception('The "audio parameter provided is not a np.ndarray.')
    
    # TODO: This won't work
    AudioFileClip(audio)


from yta_multimedia.video.edition.effect.moviepy.moviepy_effect import MoviepyEffect
from moviepy.editor import vfx, VideoFileClip, CompositeVideoClip, ImageClip
from typing import Union


class ScrollMoviepyEffect(MoviepyEffect):
    """
    This effect will make the clip appear in black and
    white colors.
    """
    __MOVIEPY_EFFECT_NAME = 'scroll'
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], width: int = None, height: int = None, x_speed: int = None, y_speed: int = None, x_start: int = None, y_start: int = None):
        self.__clip = clip

        if width is None:
            width = 960

        if height is None:
            height = 540

        if x_speed is None:
            x_speed = 20

        if y_speed is None:
            y_speed = 20

        if x_start is None:
            x_start = 100

        if y_start is None:
            y_start = 100

        self.__parameters['w'] = width
        self.__parameters['h'] = height
        self.__parameters['x_speed'] = x_speed
        self.__parameters['y_speed'] = y_speed
        self.__parameters['x_start'] = x_start
        self.__parameters['y_start'] = y_start

    def get_moviepy_vfx_effect(self):
        return getattr(vfx, self.__MOVIEPY_EFFECT_NAME, None)
    
    def process_parameters(self):
        if not self.__parameters['w']:
            self.__parameters['w'] = self.__clip.w / 2 - self.__clip.w / 2
            self.__parameters['h'] = self.__clip.h / 2 - self.__clip.h / 2
            self.__parameters['x_speed'] = 5
            self.__parameters['y_speed'] = 5
            self.__parameters['x_start'] = self.__clip.w / 2 - self.__clip.w / 2
            self.__parameters['y_start'] = self.__clip.h / 2 - self.__clip.h / 2

        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        return self.__clip.fx(self.get_moviepy_vfx_effect(), **self.process_parameters()).resize(self.__clip.size)



from yta_general_utils.checker.type import variable_is_type
from yta_general_utils.file.checker import file_is_video_file
from moviepy.editor import VideoFileClip, ImageClip, CompositeVideoClip
from typing import Union
from abc import ABC, abstractmethod


class VideoEffect(ABC):
    """
    Abstract class to be inherited by all my custom effects so I can 
    control they belong to this family.

    A video effect is an effect that is customly made by using 
    personal modifications, calculations, involving maybe some
    image manipulation, etc.
    """
    def __init__(self, video: Union[VideoFileClip, ImageClip, CompositeVideoClip, str]):
        if not video:
            raise Exception('No "video" provided.')
        
        if variable_is_type(video, str):
            if not file_is_video_file(video):
                raise Exception('Provided "video" is not a valid video file.')
            
            video = VideoFileClip(video)

        self.video = video

    @abstractmethod
    def apply(self):
        pass


import numpy as np
import math


"""
        In place movements (static) effect position functions below
"""
def circular_movement(t, x, y, radius: int = 200, cicle_time: float = 2):
    """
    Returns the (x, y) position tuple for the moviepy '.set_position()' effect,
    for each 't' provided, that will make the element move in circles with the
    provided 'radius'. The 'radius' parameter is the distance between the origin
    and the path the clip will follow. The 'cicle_time' is the time (in seconds)
    needed for a complete circle to be completed by the movement.

    If you provide the video duration as 'cicle_time', the video will make only
    one whole circle
    """
    # TODO: Do checkings
    return x + radius * np.cos((t / cicle_time) * 2 * math.pi), y + radius * np.sin((t / cicle_time) * 2 * math.pi)


"""
        With displacement (move) effect position functions below
"""
def linear_movement(t, initial_position: tuple, final_position: tuple, duration: float, fps: int):
    """
    Returns the (x, y) position tuple for the moviepy '.set_position()' method,
    for each 't' provided', that will make the element move doing a linear
    movement from the 'initial_position' to the 'final_position' in the given
    'duration' time.
    """
    # TODO: Do checkings
    progress = __get_current_movement_progress(t, duration, fps)
    x = initial_position[0] + progress * (final_position[0] - initial_position[0])
    y = initial_position[1] + progress * (final_position[1] - initial_position[1])

    return x, y

def sinusoidal_movement(t, initial_position: tuple, final_position: tuple, duration, fps: int):
    """
    Returns the (x, y) position tuple for the moviepy '.set_position()' method,
    for each 't' provided, that will make the element move doing a sinusoidal
    movement from the 'initial_position' to the 'final_position' in the provided
    'duration' time.
    """
    # TODO: Do checkings
    progress = __get_current_movement_progress(t, duration, fps)
    amplitude = 100  # amplitude of the sine wave
    frequency = 2  # frequency of the sine wave
    x = initial_position[0] + (final_position[0] - initial_position[0]) * progress
    y = initial_position[1] + (final_position[1] - initial_position[1]) * progress + amplitude * np.sin(2 * np.pi * frequency * progress)
    return (x, y)

def wave_movement(t, initial_position: tuple, final_position: tuple, duration, fps: int):
    """
    Returns the (x, y) position tuple for the moviepy '.set_position()'
    method, for each 't' provided, that will make the element go to the
    'initial_position' to the 'final_position' by doing two arcs (a 
    wave), one top and one bottom.
    """
    # TODO: Do checkings
    progress = __get_current_movement_progress(t, duration, fps)
    x = initial_position[0] + (final_position[0] - initial_position[0]) * progress
    y_start = initial_position[1]
    y_end = final_position[1]
    
    # Parameters for the sinusoidal arc
    amplitude = 200  # Maximum height of the arc
    frequency = 2 * np.pi / (final_position[0] - initial_position[0])  # Controls the frequency of the sine wave
    offset = (final_position[0] - initial_position[0]) / 2  # Center of the arc
    
    # Sinusoidal curve: y = amplitude * sin(frequency * (x - offset)) + average_y
    average_y = (y_start + y_end) / 2
    y = amplitude * np.sin(frequency * (x - initial_position[0] - offset)) + average_y
    
    return (x, y)

def arc_movement(t, initial_position: tuple, final_position: tuple, duration, fps: int, arc_is_bottom = False, max_height: int = 300):
    """
    Returns the (x, y) position tuple for the moviepy '.set_position()'
    method, for each 't' provided, that will make the element go to the
    'initial_position' to the 'final_position' by doing one single 
    arc movement. The movement will be an arc above, or bottom if the
    'arc_is_bottom' parameter is set as True. The 'max_height' parameter
    is to set the maximum height of the arc.
    """
    # TODO: Do checkings
    progress = __get_current_movement_progress(t, duration, fps)
    x = initial_position[0] + (final_position[0] - initial_position[0]) * progress
    
    # Calculate the midpoint and the maximum y value
    midpoint_x = (initial_position[0] + final_position[0]) / 2
    midpoint_y = (initial_position[1] + final_position[1]) / 2
    if arc_is_bottom:
        midpoint_y += max_height
    else:
        midpoint_y -= max_height

    # Calculate the arc movement:   y = a(x - h)^2 + k
    if arc_is_bottom:
        a = 4 * max_height / ((final_position[0] - initial_position[0])**2)
        y = -a * (x - midpoint_x)**2 + midpoint_y
    else:
        a = -4 * max_height / ((final_position[0] - initial_position[0])**2)
        y = -a * (x - midpoint_x)**2 + midpoint_y
    
    return (x, y)
    
def __get_current_movement_progress(t, duration: float, fps: int = 60):
    """
    Returns the current movement progress measured by the division
    't / (duration - 1 / fps)', that is the progress of the movement
    animation at the current moment 't'.
    """
    if not t and t != 0:
        raise Exception('No "t" provided.')
    
    if not duration:
        raise Exception('No "duration" provided.')
    
    if not fps:
        raise Exception('No "fps" provided.')

    return t / (duration - 1 / fps)


from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, concatenate_videoclips
from yta_multimedia.video.edition.effect.moviepy.fade_in_moviepy_effect import FadeInMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.fade_out_moviepy_effect import FadeOutMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.moviepy_effect import MoviepyEffect
from typing import Union


class BlinkMoviepyEffect(MoviepyEffect):
    """
    This method makes the provided video blink, that is a composition of
    a FadeOut and a FadeIn consecutively to build this effect. The duration
    will be the whole clip duration. The FadeIn will last the half of the
    clip duration and the FadeOut the other half.

    The 'color' parameter is the color you want for the blink effect as the
    background color. The default value is black ([0, 0, 0]).
    """
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], color = None):
        self.__clip = clip

        if color is None:
            color = [0, 0, 0]

        self.__parameters['final_color'] = color

    def process_parameters(self):
        # TODO: Is this acutally needed (?) Is doing nothing
        # TODO: Check color is a valid value maybe (?)
        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        if not self.__clip:
            return None
        
        self.process_parameters()
        
        half_duration = self.__clip.duration / 2
        self.__clip = concatenate_videoclips([
            FadeOutMoviepyEffect(self.__clip.subclip(0, half_duration), duration = half_duration, color = self.__parameters['final_color']).apply(),
            FadeInMoviepyEffect(self.__clip.subclip(half_duration, self.__clip.duration), duration = half_duration, color = self.__parameters['final_color']).apply(),
        ])

        return self.__clip


from yta_multimedia.video.frames import get_frame_from_video_by_frame_number
from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import ImageClip, concatenate_videoclips


class StopMotionVideoEffect(VideoEffect):
    """
    Creates a Stop Motion effect in the provided video by dropping the frames
    per second but maintaining the original frames ratio.
    """
    def apply(self):
        # This value is this one by default by now
        FRAMES_TO_JUMP = 5

        clips = []
        for frame_number in range((int) (self.video.fps * self.video.duration)):
            if frame_number % FRAMES_TO_JUMP == 0:
                frame = get_frame_from_video_by_frame_number(self.video, frame_number)
                clips.append(ImageClip(frame, duration = FRAMES_TO_JUMP / self.video.fps).set_fps(self.video.fps))

        return concatenate_videoclips(clips).set_audio(self.video.audio).set_fps(self.video.fps)


    


from yta_multimedia.video.edition.effect.moviepy.moviepy_effect import MoviepyEffect
from yta_general_utils.temp import create_temp_filename, create_custom_temp_filename
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, AudioFileClip, ImageSequenceClip
from typing import Union
from pydub import AudioSegment


class ReversedMoviepyEffect(MoviepyEffect):
    """
    This method creates a new one but in reversa, also with the sound reversed.

    It doesn't use the 'mirror_time' effect because it fails. Instead, it saves
    each frame of the video and builds a new video using them in reverse order.
    It also uses the original audio an reverses it in the new generated video.
    """
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip]):
        self.__clip = clip

    def process_parameters(self):
        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        if not self.__clip:
            return None
        
        self.process_parameters()
        
        # TODO: Try to do this in memory, not writing files
        AUDIO_FILE = create_temp_filename('tmp_audio.mp3')
        REVERSED_AUDIO_FILE = create_temp_filename('tmp_reversed_audio.mp3')
        frames_array = []
        for frame in self.__clip.iter_frames():
            frame_name = create_custom_temp_filename('frame_' + str(len(frames_array)) + '.png')
            frames_array.append(frame_name)
        self.__clip.write_images_sequence(create_custom_temp_filename('') + 'frame_%01d.png', logger = 'bar')

        # Reverse audio (I set it manually to fps because others
        # values are failing)
        self.__clip.audio.write_audiofile(AUDIO_FILE, fps = 44100)
        original = AudioSegment.from_mp3(AUDIO_FILE)
        original.reverse().export(REVERSED_AUDIO_FILE)
        reversed_audio = AudioFileClip(REVERSED_AUDIO_FILE)

        # [ 2 ]   Rebuild the video, but in reverse, from last frame to first one
        frames_array = frames_array[::-1]
        self.__clip = ImageSequenceClip(frames_array, fps = self.__clip.fps).set_audio(reversed_audio)

        return self.__clip


from yta_general_utils.file.checker import file_is_video_file
from yta_general_utils.file.filename import file_is_type, FileType
from moviepy.editor import VideoFileClip, CompositeVideoClip, ColorClip, ImageClip
from typing import Union


def video_to_moviepy_video(video: Union[str, VideoFileClip, CompositeVideoClip, ColorClip, ImageClip]):
    """
    This method is a helper to turn the provided 'video' to a moviepy
    video type. If it is any of the moviepy video types specified in
    method declaration, it will be returned like that. If not, it will
    be load as a VideoFileClip if possible, or will raise an Exception
    if not.
    """
    if not video:
        raise Exception('No "video" provided.')
    
    if not isinstance(video, str) and not isinstance(video, VideoFileClip) and not isinstance(video, CompositeVideoClip) and not isinstance(video, ColorClip) and not isinstance(video, ImageClip):
        raise Exception('The "video" parameter provided is not a valid type. Check valid types in method declaration.')
    
    if isinstance(video, str):
        if not filename_is_type(video, FileType.VIDEO):
            raise Exception('The "video" parameter provided is not a valid video filename.')
        
        if not file_is_video_file(video):
            raise Exception('The "video" parameter is not a valid video file.')
        
        video = VideoFileClip

    return video



from yta_general_utils.checker.type import variable_is_type
from yta_general_utils.image.converter import numpy_image_to_pil
from PIL import Image
from typing import Union, Any

import numpy as np
import cv2


# TODO: Make 'image: Union[str, Image.Image, np.ndarray]' work please
def resize_image_scaling(image: Union[str, Any], width: int, height: int, output_filename: Union[str, None] = None):
    """
    Resizes the provided 'image' (if valid) to the provided dimensions
    ('width', 'height') scaling it (losing not the aspect ratio). The
    image will be stored locally as 'output_filename' if provided and
    will be returned once it's been scalled, in any case.

    TODO: Please, test and check this method
    """
    # TODO: This method needs to be checked properly
    if not image:
        raise Exception('No "image" provided.')

    if not width:
        raise Exception('No "width" provided.')
    
    if not height:
        raise Exception('No "height" provided.')

    if variable_is_type(image, str):
        # TODO: Check if image is actually an image
        image = Image.open(image)

    if variable_is_type(image, np.ndarray):
        image = numpy_image_to_pil(image)

    # Lets resize it
    image_width, image_height = image.size

    if image_width == width and image_height == height:
        return image.save(output_filename)

    aspect_ratio = image_width / image_height
    if aspect_ratio > (width / height):
        # Image is very horizontal, so width changes faster, we need to focus on height
        factor = (height * 100 / image_height) / 100
        image_width = int(image_width * factor)
        image_height = height
    else:
        # Image is very vertical, so height changes faster, we need to focus on width
        factor = (width * 100 / image_width) / 100
        image_width = 1920
        image_height = int(image_height * factor)
    image = image.resize((image_width, image_height))

    # We will crop form the center to edges
    left = 0
    right = width
    if image_width > width:
        # If it is 1960 => leave [0, 20], get [20, 1940], leave [1940, 1960]
        margin = int((image_width - width) / 2)
        left = 0 + margin
        right = image_width - margin
        # We make and adjustment if some pixel left
        while (right - left) > width:
            right -= 1
        while (right - left) < width:
            if left > 0:
                left -= 1
            else:
                right += 1

    top = 0
    bottom = height
    if image_height > height:
        # If it is 1140 => leave [0, 30], get [30, 1110], leave [1110, 1140]
        margin = int((image_height - height) / 2)
        top = 0 + margin
        bottom = image_height - margin
        # We make and adjustment if some pixel left
        while (bottom - top) > height:
            bottom -= 1
        while (bottom - top) < height:
            if top > 0:
                top -= 1
            else:
                bottom += 1

    # Image that is 1920x1080 and is the center of the original image
    image = image.crop((left, top, right, bottom))

    if output_filename:
        image.save(output_filename)

    return image

def resize_image_without_scaling(image: Union[str, Any], width: int, height: int, output_filename: Union[str, None] = None):
    """
    Resizes the provided 'image' (if valid) to the provided dimensions
    ('width', 'height') guaranteen not that the aspect ratio is kept. 
    The image will be stored locally as 'output_filename' if provided
    and will be returned once it's been scalled, in any case.

    TODO: Please, test and check this method
    """
    # TODO: This method needs to be checked properly
    if not image:
        raise Exception('No "image" provided.')

    if not width:
        raise Exception('No "width" provided.')
    
    if not height:
        raise Exception('No "height" provided.')

    if variable_is_type(image, str):
        # TODO: Check if image is actually an image
        image = Image.open(image)

    if variable_is_type(image, np.ndarray):
        image = numpy_image_to_pil(image)

    # From here: https://stackoverflow.com/a/14140796
    image = image.convert('RGB')
    image = np.array(image) # To cv2
    image = image[:, :, ::-1].copy() # Convert RGB to BGR

    #image = cv2.imread(image_filename)
    resized_image = cv2.resize(image, dsize = (width, height), interpolation = cv2.INTER_CUBIC)

    if output_filename:
        cv2.imwrite(output_filename, resized_image)

    return resized_image


# TODO: Remove these 2 methods below if the ones above are working, because
# they are more general (using different types or images)

def resize_image_file_scaling(image_filename: str, width: int, height: int, output_filename = None):
    """
    Resizes the provided 'image_filename' to the provided 'width' and 'height' keeping the
    aspect ratio. This method enlarges the image to fit the desired size and then makes a 
    crop to obtain that size from the center of the resized image. If 'output_filename' is
    provided, the image is saved locally with that name. If not, it is only returned
    """
    if not image_filename:
        return None
    
    # TODO: Do some checkings on width and height (?)

    # TODO: This is not working well when '' passed
    if output_filename is not None and not output_filename:
        return None

    image = Image.open(image_filename)
    image_width, image_height = image.size

    if image_width == width and image_height == height:
        return image.save(output_filename)

    aspect_ratio = image_width / image_height
    if aspect_ratio > (width / height):
        # Image is very horizontal, so width changes faster, we need to focus on height
        factor = (height * 100 / image_height) / 100
        image_width = int(image_width * factor)
        image_height = height
    else:
        # Image is very vertical, so height changes faster, we need to focus on width
        factor = (width * 100 / image_width) / 100
        image_width = 1920
        image_height = int(image_height * factor)
    image = image.resize((image_width, image_height))

    # We will crop form the center to edges
    left = 0
    right = width
    top = 0
    bottom = height
    if image_width > width:
        # If it is 1960 => leave [0, 20], get [20, 1940], leave [1940, 1960]
        margin = int((image_width - width) / 2)
        left = 0 + margin
        right = image_width - margin
        # We make and adjustment if some pixel left
        while (right - left) > width:
            right -= 1
        while (right - left) < width:
            if left > 0:
                left -= 1
            else:
                right += 1
    if image_height > height:
        # If it is 1140 => leave [0, 30], get [30, 1110], leave [1110, 1140]
        margin = int((image_height - height) / 2)
        top = 0 + margin
        bottom = image_height - margin
        # We make and adjustment if some pixel left
        while (bottom - top) > height:
            bottom -= 1
        while (bottom - top) < height:
            if top > 0:
                top -= 1
            else:
                bottom += 1

    image = image.crop((left, top, right, bottom))
    # Image that is 1920x1080 and is the center of the original image
    if output_filename:
        image.save(output_filename)

    return image

def resize_image_file_without_scaling(image_filename: str, width: int = 1920, height: int = 1080):
    """
    This method gets an image, resizes it and overwrites the original one.

    TODO: This method need work.
    """
    if not image_filename:
        return None
    
    # TODO: Do some checkings on width and height (?)

    # TODO: We resize it simply, we don't care about scale
    image = cv2.imread(image_filename)
    resized_image = cv2.resize(image, dsize = (width, height), interpolation = cv2.INTER_CUBIC)
    cv2.imwrite(image_filename, resized_image)


from yta_multimedia.greenscreen.classes.greenscreen_area_details import GreenscreenAreaDetails
from yta_general_utils.downloader.google_drive import get_id_from_google_drive_url
from yta_general_utils.checker.url import is_google_drive_url
from yta_multimedia.resources import get_resource
from yta_multimedia.greenscreen.enum.greenscreen_type import GreenscreenType
from yta_multimedia.greenscreen.utils import GREENSCREENS_FOLDER
from typing import List


class GreenscreenDetails:
    """
    This class represents a greenscreen image or video
    resource, its information and the greenscreen areas
    that it contains.

    @param
        **greenscreen_areas**
        An array of _GreenscreenAreaDetails_ objects that
        represent the different greenscreen that there are
        inside this greenscreen resource.

    @param
        **filename_or_google_drive_url**
        The resource file filename or Google Drive url.
    """
    greenscreen_areas = []
    filename_or_google_drive_url = None
    type = None

    def __init__(self, greenscreen_areas: List[GreenscreenAreaDetails] = [], filename_or_google_drive_url: str = None, type: GreenscreenType = GreenscreenType.IMAGE):
        # TODO: Implement checkings please
        self.greenscreen_areas = greenscreen_areas
        self.filename_or_google_drive_url = filename_or_google_drive_url
        self.type = type

    def get_filename(self):
        """
        This method will return the video or image resource you need
        to use the greenscreen, that will be stored locally (and 
        maybe downloaded from Google Drive if it is not available
        in local storage yet).
        """
        # TODO: Change this behaviour and be more strict
        # and make more fields if needed, please
        filename = self.filename_or_google_drive_url

        if is_google_drive_url(self.filename_or_google_drive_url):
            google_drive_id = get_id_from_google_drive_url(self.filename_or_google_drive_url)
            folder = GREENSCREENS_FOLDER + google_drive_id + '/'

            filename = folder + 'greenscreen.mp4'
            if self.type == GreenscreenType.IMAGE:
                filename = folder + 'greenscreen.png'

            filename = get_resource(self.filename_or_google_drive_url, filename)

        return filename


from yta_multimedia.video.edition.duration import set_video_duration
from yta_general_utils.file.checker import file_is_video_file
from moviepy.editor import ColorClip, VideoFileClip, CompositeVideoClip, ImageClip
from typing import Union


class BasePositionMoviepyEffect:
    """
    Class created to test position effects and building objects
    to simplify their use in our system.
    """
    video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip] = None

    def __init__(self, video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip]):
        # TODO: This could be an Image that we want to make clip
        if not video:
            raise Exception('No "video" provided.')
    
        if isinstance(video, str):
            if not file_is_video_file:
                raise Exception('Provided "video" is not a valid video file.')
            
            video = VideoFileClip(video)

        self.video = video

    # TODO: Apply the effect within a black background clip of 1920x1080
    def apply(self):
        """
        Applies the effect to the 'video' provided when initializing this
        effect class, and puts the video over a static black background
        image of 1920x1080.
        """
        background_video = ColorClip((1920, 1080), [0, 0, 0], duration = self.video.duration)

        return self.apply_over_video(background_video)

    def process_background_video(self, background_video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        This method will process the provided 'background_video' to check that it is
        valid and also to change its duration to fit the 'self.video' given when
        initializing this effect object. It will shorten the video if its duration is
        greater than the 'self.video', or will loop it to reach the 'self.video'
        duration.

        This method returns, if everything is ok, the new 'background_video' with its
        new duration.
        """
        if not background_video:
            raise Exception('No "background_video" provided.')
    
        if isinstance(background_video, str):
            if not file_is_video_file:
                raise Exception('Provided "background_video" is not a valid video file.')
            
            background_video = VideoFileClip(background_video)

        background_video = set_video_duration(background_video, self.video.duration)

        return background_video
    
    def apply_over_video(self, background_video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        Applies the effect on the video used when instantiating the
        effect, but applies the effect by placing it over the 
        'background_video' provided in this method (the 
        'background_video' will act as a background video for the 
        effect applied on the initial video).

        This method will set the video used when instantiating the
        effect as the most important, and its duration will be 
        considered as that. If the 'background_video' provided 
        has a duration lower than the original video, we will
        loop it to reach that duration. If the video is shorter
        than the 'background_video', we will crop the last one
        to fit the original video duration.
        """
        pass


def colouring_book_sketch(element: str):
    """
    Returns a prompt to generate a colouring book or sketch
    image. This means the shape of the provided 'element' 
    with black borders, white filling and a chroma key green
    background to be able to remove the background of the
    image easily.
    """
    return 'Create a simple, child-friendly coloring page featuring a ' + element + '. The ' + element + ' should have crisp, clean black outlines with no shading or gradients. The interior of the ' + element + ' should be filled with pure white. Ensure that the black lines are distinct and well-defined. The background of the image should be a solid, intense green, like chroma key green, with no additional details or elements.'